from .bot import Bot
from ..direction import Direction
from .algs import BFS

class ExtrovertBot(Bot):
    '''
    ExtrovertBot find the position in the map where there are the most amount
    of people.
    '''

    ''' 
    We add this value to the population of the neighborhood tiles when
    deciding whether or to the change the destination tile to a different
    adjacent tile. This is added just in case our ExtrovertBot changes
    directions too often and ultimately does not move. Increase this
    value to ensure the bot moves more, decrese this value so that
    its more purely an extrovert.
    Example:
    Set this to zero if you always want to change your destination tile
    to the neighboring location with the most people.
    Example:
    Set this to one if one of the neighboring locations needs two more
    people on it for this bot the change their direction.
    '''
    REORIENT_COST = 2

    def __init__(self, board, loc, speed, id):
        super().__init__(board, loc, speed, id)
        self.queue = []

    def find_big_brain(self):
        '''
        Finds the tile with the most people.
        '''
        max_pop = 0
        max_point = None

        for i in range(self.board.x_dim()):
            for j in range(self.board.y_dim()):
                if self.board.get((i,j)).get_num_bots() > max_pop:
                    max_pop = self.board.get((i,j)).get_num_bots()
                    max_point = (i, j)
        return max_point

    
    def compute_step(self):
        dest = self.find_big_brain() 
        path = BFS(self.board, self.loc, lambda tile: tile.get_loc() == dest)
        if not self.queue or \
           self.board.get(self.queue[-1]).get_num_bots() + self.REORIENT_COST < \
           self.board.get(path[-1]).get_num_bots():
            self.queue = path
        
        if self.queue:
            if self.loc == self.queue[0]:
                self.queue.pop(0)
            if self.queue:
                first_tile = self.queue[0]
                self.new_direction = Direction.get_dir(self.loc, first_tile)




name = 'bots'
from .bot import Bot
from .jittery import JitteryBot
from .still import StillBot
from .follow import LineFollower as FollowBot
from .extrovert import ExtrovertBot
from .realistic import RealisticBot
from .random import RandomBot


from .bot import Bot
from ..direction import Direction
from ..tile import Booth
from .algs import BFS

class RealisticBot(Bot):
    """
    RealisticBot aims for the smaller companies because it has the best
    chance of getting a job from a smaller company. It will not visit the
    same company twice, because there is no point in handing a resume out
    twice so it will mark the companies it has visited. It uses a bfs to
    search for the closest small company and move towards that direction.
    It should find a company and retrieve the path.
    """
    def __init__(self, board, loc, speed, id):
        super().__init__(board, loc, speed, id)
        self.queue = []
        self.to_visit = "S"
        self.visited = set()
    
    def update_to_visit(self):
        if self.to_visit == "S":
            self.to_visit = "M"
        elif self.to_visit == "M":
            self.to_visit = "L"
        else:
            self.to_visit = "S"
    
    def compute_step(self):
        if not self.is_in_line():
            src = self.loc
            self.queue = BFS(self.board, src, 
                             lambda tile: \
                             self.board.booths.get(tile.get_booth()) and \
                             self.board.booths.get(tile.get_booth()).get_size() == self.to_visit \
                             and tile.get_booth() not in self.visited)
            while self.queue is None:
                self.update_to_visit()
                self.visited = set()
                self.queue = BFS(self.board, src, 
                                 lambda tile: \
                                 self.board.booths.get(tile.get_booth()) and \
                                 self.board.booths.get(tile.get_booth()).get_size() == self.to_visit \
                                 and tile.get_booth() not in self.visited)
            self.queue = BFS(self.board, src, lambda tile: tile.get_line() == self.board.get(self.queue[-1]).get_booth())

        if self.queue:
            if self.loc == self.queue[0]:
                self.queue.pop(0)
            if self.queue:
                first_tile = self.queue[0]
                self.new_direction = Direction.get_dir(self.loc, first_tile)
            else:
                self.new_direction = Direction.ENTER
                self.visited.add(self.board.get(self.loc).get_line())



from .bot import Bot
from ..direction import Direction
import random

LINE_PROB = 0.6

class JitteryBot(Bot):
    def __init__(self, board, loc, speed, id):
        super().__init__(board, loc, speed, id)
        self.next_loc = self.loc

        self.directions = [Direction.LEFT, Direction.UP, Direction.RIGHT, Direction.DOWN]
    
    def compute_step(self):
        if self.in_line:
            self.next_loc = self.loc
        elif self.loc == self.next_loc:
            rand_num = random.uniform(0, 1)
            curr_tile = self.board.get(self.loc)
            if curr_tile.is_end_of_line() and rand_num <= LINE_PROB:
                self.new_direction = Direction.ENTER
            else: 
                random.shuffle(self.directions)
                for direction in self.directions:
                    if self._is_valid(direction):
                        self.new_direction = direction
                        break
                self.next_loc = self.new_direction.get_loc(self.loc)



from .bot import Bot
from ..direction import Direction
from .algs import BFS
import random

class RandomBot(Bot):
    '''
    Defines a bot which goes to a random company of the given size.
    The possible sizes are None, S, M, L. If the size is None then all companies
    are allowed, otherwise only companies of the same size are allowed.
    After exiting a line for a company, the bot will again randomly select a company.
    '''
    def __init__(self, board, loc, speed, id, size=None):
        super().__init__(board, loc, speed, id)
        self.queue = []
        self.choices = None
        self.choices = [name for name in self.board.booths if self.board.booths[name].get_size() == size]
        if len(self.choices) == 0:
            self.choices = [name for name in self.board.booths]
        self.booth = None
    
    def compute_step(self):
        if not self.is_in_line():
            src = self.loc
            if not self.queue:
                self.booth = random.choice(self.choices)
                self.queue = BFS(self.board, src, lambda tile: tile.is_end_of_line() and tile.get_line() == self.booth)

        if self.queue:
            if self.loc == self.queue[0]:
                self.queue.pop(0)
            if self.queue:
                first_tile = self.queue[0]
                self.new_direction = Direction.get_dir(self.loc, first_tile)
            else:
                self.new_direction = Direction.ENTER



from ...direction import Direction
from random import shuffle

def BFS(board, src, p):
    dirs = [Direction.LEFT, Direction.UP, Direction.RIGHT, Direction.DOWN]

    queue = [[src]]
    visited = set()

    while queue:
        path = queue.pop(0)
        node = path[-1]

        tile = board.get(node)

        if p(tile):
            return path
        if node not in visited:
            visited.add(node)
        else:
            continue

        shuffle(dirs)
        locs = [direction.get_loc(node) for direction in dirs]
        for loc in locs:
            tile = board.get(loc)
            if tile != None and p(tile):
                return list(path) + [loc]
            if tile != None and tile.get_booth() is None:
                new_path = list(path)
                new_path.append(loc)
                queue.append(new_path)
